/**
 * @file Implémentation des requêtes GraphQL.
 * @author akka vodol, ofacklam
 * @memberof GraphQL
*/

import { AuthenticationError } from "apollo-server-core";
import { Context } from "./typeDefs/queries";
import { User } from "./resolvers/users";
import { Group, SimpleGroup, MetaGroup } from "./resolvers/groups";
import { Announcement, Event, PrivatePost, Question, Answer, Message } from "./resolvers/messages";
import { UserJoinGroup, GroupJoinMetagroup, GroupCoauthorEvent, Request } from "./resolvers/requests";
import { GroupCollection } from "./connectors/tools";

/*
    Le tag @rights est la gestion des autorisations.

    Le système GraphQL est pensé comme l'interface par laquelle les utilisateurs
    intéragissent avec sigma, les graphismes en moins.
    Le client peut envoyer tout type de requête. C'est au niveau des resolvers
    que les permissions sont gérées. D'où le @rights

    Commençons par un rappel sur le fonctionnement des droits.
    Chaque utilisateur a un certain niveau de droit sur chaque groupe. Ce niveau de droit indique
    ce qu'il a le droit de savoir et de faire. Chaque niveau est inclus dans les niveaus supérieur.
    Les différents niveaux sont :
    none - aucun droit
    viewer : l'utilisateur a visibilité sur le groupe. Il sait que le groupe existe, et a accès à un certain nombre d'infos.
    member : l'utilisateur est membre du groupe
    speaker : l'utilisateur peut parler au nom du groupe. Il a le droit de publier des annonces et d'organiser des évènements
    admin : l'utilisateur a tous les droits sur le groupe

    Certaines fonctions de connectors effectuent des vérifications d'authorisations avant
    de renvoyer une réponse, d'autres non. Pour être sur qu'on ne renvoie jamais de réponse
    sans avoir au préalable éffectué les bonnes vérifications, chaque fonction possède dans sa
    description un attribut droit, qui décrit les droits que fournit cette fonction.

    La valeur de @rights peut être :
    super - la fonction n'effectue aucune véri-fication, et renvoie le resultat demandé
    admin( groupUID ) - la fonction ne fait que ce qu'un admin du groupe indiqué aurait le droit de faire
    speaker( groupUID ), member( groupUID ), veiwer( groupUID ) - même chose
    user - la fonction ne fait que ce que l'utiliateur a le droit de faire (vérifications via l'argument user)

    La procédure a suivre est la suivante : quand une fonction possède un certain niveau de droit,
    elle ne peut appeler une fonction possédant un niveau de droit plus large que si
    1 ) on a au préalable vérifié que l'utilisateur possédait effectivement ces droits.
    ou
    2 ) on s'est assuré que l'opération effectuée par cet appel particulier de la fonction était dans les droits
    de l'utilisateur

    Les resolvers de base de mutation et query ont des droits user.

    Les fonctions qui ne modifient pas la BDD et ne renvoient pas de données sur la BDD n'ont pas de rights.
*/

/**
 * Résolveurs des différentes requêtes GraphQL
 */

export const resolvers = {
    Query: {

        // User queries de base
        // @rights connectedOrOnplatal
        user: async function (root, args, context: Context): Promise<User> {
            if(context.models.auth.isConnectedOrOnplatal()) {
                return context.models.user.getUser(args.uid);
            }
            throw new AuthenticationError("Not connected or on-platal");
        },

        // Group queries de base
        // @rights connectedOrOnplatal
        group: async function (root, args, context: Context): Promise<Group> {
            if (context.models.auth.isConnectedOrOnplatal()) {
                return context.models.group.getGroup(args.gid);
            }
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights connectedOrOnplatal
        simpleGroup: async function (obj, args, context: Context): Promise<SimpleGroup> {
            if (context.models.auth.isConnectedOrOnplatal()) {
                return context.models.group.getSimpleGroup(args.gid);
            }
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights connectedOrOnplatal
        metaGroup: async function (obj, args, context: Context): Promise<MetaGroup> {
            if (context.models.auth.isConnectedOrOnplatal()) {
                return context.models.group.getMetaGroup(args.gid);
            }
            throw new AuthenticationError("Not connected or on-platal");
        },

        // Message queries de base
        //message -> not implemented -> message(mid: ID!): Message

        // @rights membre d'un groupe author ou d'un groupe recipient
        announcement: async function (root, args, context: Context): Promise<Announcement> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.message.getAnnouncement(args.mid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights membre d'un groupe author ou d'un groupe recipient
        event: async function (root, args, context: Context): Promise<Event> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.message.getEvent(args.mid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights membre du groupe recipient
        privatePost: async function (root, args, context: Context): Promise<PrivatePost> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.message.getPrivatePost(args.mid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights viewer du groupe recipient
        question: async function (root, args, context: Context): Promise<Question> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.message.getQuestion(args.mid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights viewer du groupe author
        answer: async function (root, args, context: Context): Promise<Answer> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.message.getAnswer(args.mid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // Request queries de base
        //request -> not implemented -> request(rid: ID!): Request

        // @rights admin du groupe destinaire ou le user émetteur
        userJoinGroupRequest: async function (root, args, context: Context): Promise<UserJoinGroup> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.request.getUserJoinGroupRequest(args.rid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights admin du groupe émetteur ou destinataire
        groupJoinMetagroupRequest: async function (root, args, context: Context): Promise<GroupJoinMetagroup> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.request.getGroupJoinMetagroupRequest(args.rid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // @rights admin du groupe émetteur ou destinataire
        groupCoauthorEventRequest: async function (root, args, context: Context): Promise<GroupCoauthorEvent> {
            //TODO : verifier les autorisations
            throw "Not implemented";
            return context.models.request.getGroupCoauthorEventRequest(args.rid);
            throw new AuthenticationError("Not connected or on-platal");
        },

        // Tous les Messages visibles par un utilisateur(dont le uid, et donc les autorisations, est passé par context)
        // @rights member of groups
        allMessages: async function (root, args, context: Context): Promise<Message[]> {
            let groups = context.models.auth.groupsMember();
            return context.models.message.getAllMessages(groups);
        },

        // @rights member of groups
        allAnnouncements: async function (root, args, context: Context): Promise<Announcement[]> {
            let groups = context.models.auth.groupsMember();
            return context.models.message.getAllAnnouncements(groups);
        },

        // @rights member of groups
        allEvents: async function (root, args, context: Context): Promise<Event[]> {
            let groups = context.models.auth.groupsMember();
            return context.models.message.getAllEvents(groups);
        },

        // @rights member of groups
        allPrivatePosts: async function (root, args, context: Context): Promise<PrivatePost[]> {
            let groups = context.models.auth.groupsMember();
            return context.models.message.getAllPrivatePosts(groups);
        },

        // Tous les Groupes visibles par un utilisateur.
        // Correspondrait au sous - champ "viewerOf" de User, volontairement non - défini comme tel.Tous les autres cas de figure sont couverts par les sous - champs "<permission>Of" de User
        // @rights viewer of groups
        allGroups: async function (root, args, context: Context): Promise<Group[]> {
            let visibleGroupCollection = context.models.auth.groupsViewer();
            return context.models.group.getAllGroupsByCollection(visibleGroupCollection);
        },

        // @rights viewer of groups
        allSimpleGroups: async function (root, args, context: Context): Promise<SimpleGroup[]> {
            let visibleGroupCollection = context.models.auth.groupsViewer();
            return context.models.group.getAllSimpleGroups(visibleGroupCollection.simpleGroups);
        },

        // Toutes les Requests auxquelles un groupe doit répondre
        // @rights admin
        requestsToGroup: async function (root, args, context: Context): Promise<Request[]> {
            if(context.models.auth.isAdmin(args.gid)) {
                return context.models.request.getRequestsToGroup(args.gid);
            }
            throw new AuthenticationError("Not an admin");
        },

        // @rights admin
        userJoinGroupRequestsToGroup: async function (root, args, context: Context): Promise<UserJoinGroup[]> {
            if (context.models.auth.isAdmin(args.gid)) {
                return context.models.request.getUserJoinGroupRequestsToGroup(args.gid);
            }
            throw new AuthenticationError("Not an admin");
        },

        // @rights admin
        groupJoinMetagroupRequestsToGroup: async function (root, args, context: Context): Promise<GroupJoinMetagroup[]> {
            if (context.models.auth.isAdmin(args.gid)) {
                return context.models.request.getGroupJoinMetagroupRequestsToGroup(args.gid);
            }
            throw new AuthenticationError("Not an admin");
        },

        // @rights admin
        groupCoauthorEventRequestsToGroup: async function (root, args, context: Context): Promise<GroupCoauthorEvent[]> {
            if (context.models.auth.isAdmin(args.gid)) {
                return context.models.request.getGroupCoauthorEventRequestsToGroup(args.gid);
            }
            throw new AuthenticationError("Not an admin");
        },

        // TOL
        // @rights connectedOrOnplatal
        searchTOL: async function (root, args, context: Context): Promise<User[]> {
            if (context.models.auth.isConnectedOrOnplatal()) {
                return context.models.user.searchTOL(args);
            }
            throw new AuthenticationError("Not connected or on-platal");
        }
    },

    Mutation: {

        // Superviser mutations

        takeAdminRights : async function(obj, args, context) : Promise<boolean>{
            const justification = await auth.isSupervisor(context.user, args.from);
            if(justification){
                return conn.takeAdminRights(context.user, args.from, justification);
            }else{
                return false;
            }
        },

        releaseAdminRights : async function(obj, args, context){
            await conn.releaseAdminRights(context.user, args.from);
        },

        // Admin mutations

        createSubgroup: async function (obj, args, context){
            throw new Error('Not implemented');
        },
    }
};
